#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "dialitd.h"
#include "commands.h"
#include "tree.h"

static command_t commands[]=
{
    {CMD_NO_ARGS, command_shut, "SHUT"},
    {CMD_NO_ARGS, command_quit, "QUIT"},
    {CMD_NO_ARGS, command_help, "HELP"},
    {CMD_NO_ARGS, command_scan, "SCAN"},
    {CMD_NO_ARGS, command_list, "LIST"},
    {CMD_NO_ARGS, command_stat, "STAT"},
    {CMD_OPT_ARGS,command_stop, "STOP"},
    {CMD_HAS_ARGS,command_dial, "DIAL"},
    {0, NULL, NULL}
};

int command(config_t * config,char * cmnd_ln,int client_number)
{
    int loop;
    int match;

    /* search command list for match */
    for(loop=0;commands[loop].proc != NULL;loop++)
    {
	match=cmp_command(cmnd_ln,commands[loop].type,commands[loop].cmnd);

	switch(match)
	{
	case 0: /* good match */
	    return (commands[loop].proc)(config,cmnd_ln,client_number);
	case 2: /* missing arguments */
	    return command_missing_args(config,cmnd_ln,client_number);
	}
    }

    /* no matches in list so call unkown command handler */
    return command_unknown(config,cmnd_ln,client_number);
}


int cmp_command(char * string,int hasargs,char * cmnd)
{
    int str_len;
    int cmnd_len;

    str_len=strlen(string);
    cmnd_len=strlen(cmnd);

    /* check minimum length */
    if(str_len < cmnd_len)
	return 1; /* no match */

    /* check for exact match on non arument command */
    if((str_len == cmnd_len) &&
       (strncasecmp(string,cmnd,cmnd_len)==0))
    {

       if(hasargs == CMD_HAS_ARGS)
	   return 2; /* mandatory args and none provided */
       else
	   return 0; /* good match */
    }

    /* check for match with args */
    if((str_len > cmnd_len) &&
       (hasargs) &&
       (strncasecmp(string,cmnd,cmnd_len) == 0) &&
       (string[cmnd_len] == ' '))
	return 0; /* good match */

    return 1; /* no match */
}

/* command functions */

/* support functions for command functions */
int pppd_stop(char * interface)
{
    FILE * fpppdpid;
    char pppdpidname[1024];
    int pppdpid = 0;

    snprintf(pppdpidname,1024,"/var/run/%s.pid",interface);

    fpppdpid = fopen(pppdpidname,"r");
    if(fpppdpid==NULL)
    {
	/* pid file not openable */
	return 1;
    }

    /* read pid and exit if its not readable */
    if(fscanf(fpppdpid,"%d",&pppdpid)!=1)
    {
	fclose(fpppdpid);
	return 2;
    }

    fclose(fpppdpid);

    if(kill(pppdpid,SIGINT)<0)
    {
	/* sending sig failed - likely dead pid file */
	return 3;
    }

    /* signal sent ok*/

    return 0;
}

int pppd_call(char * provider)
{
    int childpid;
    int retstat;

    childpid=fork();
    if(childpid==0)
    {
	/* were the kid */
	execlp("pppd","pppd","call",provider,NULL);
	perror("dialitd");
	exit(1);
    }

    if(waitpid(childpid,&retstat,0)<0)
    {
	perror("dialitd");
	return 1;
    }

    if(WIFEXITED(retstat))
    {
	return WEXITSTATUS(retstat);
    }

    return 1;
}

/* unkown command trap */
int command_unknown(config_t * config,char * cmnd_ln,int cli_num)
{
    send_reply(config->client[cli_num].fd,400,"Unknown command %s\n",cmnd_ln);
    return 1;
}

/* unkown command trap */
int command_missing_args(config_t * config,char * cmnd_ln,int cli_num)
{
    send_reply(config->client[cli_num].fd,400,"Comamnd %s Requires Arguments\n",cmnd_ln);
    return 1;
}

int command_shut(config_t * config,char * cmnd,int client_number)
{
    send_reply(config->client[client_number].fd,100,"Shutting server down\n");
    config->quit = 1;
    return 0;
}

int command_quit(config_t * config,char * cmnd,int cli_num)
{
    send_reply(config->client[cli_num].fd,100,"Closing connection\n");
    close_client(config,cli_num);
    return 0;
}

int command_help(config_t * config,char * cmnd,int cli_num)
{
    send_reply(config->client[cli_num].fd,100,"Supported commands are\n");
    send_reply(config->client[cli_num].fd,101,"LIST, SHUT, QUIT, HELP, DIAL <account>, STOP [<interface>], SCAN, STAT\n");
    return 0;
}

int command_scan(config_t * config,char * cmnd,int cli_num)
{
    struct treenode * new_treeroot;
    send_reply(config->client[cli_num].fd,100,"Reseting provider list\n");
    new_treeroot = make_tree_list("/etc/ppp/peers");
    if(new_treeroot == NULL)
    {
	send_reply(config->client[cli_num].fd,101,"Reset FAILED! - aborting\n");
	return 1;
    }
    else
    {
	free_tree_list(config->treeroot);
	config->treeroot = new_treeroot;
	send_reply(config->client[cli_num].fd,101,"Reset Suceded\n");
    }

    return 0;
}


int walk_tree_list(int fd,struct treenode* node)
{
    if (node->left != NULL)
	walk_tree_list(fd,node->left);

    send_reply(fd,101,"%s\n",node->filename);

    if (node->right != NULL)
	walk_tree_list(fd,node->right);

    return 0;
}


int command_list(config_t * config,char * cmnd,int cli_num)
{
    send_reply(config->client[cli_num].fd,100,"Conections Available\n");
    walk_tree_list(config->client[cli_num].fd,config->treeroot);
    return 0;
}

int command_stat(config_t * config,char * cmnd,int cli_num)
{
    if(config->link_active)
	send_reply(config->client[cli_num].fd,100,"Conection is UP\n");
    else
	send_reply(config->client[cli_num].fd,100,"Conection is DOWN\n");
    return 0;
}

int command_dial(config_t * config,char * cmnd_ln,int cli_num)
{
    struct treenode ** link;
    char * value;

    /* skip the dial keyword */
    value=&cmnd_ln[5];

    link=find_treenode(&config->treeroot,value);
    if (*link==NULL)
    {
	send_reply(config->client[cli_num].fd,400,"Unknown account %s\n",value);
	return 1;
    }
    else
    {
	send_reply(config->client[cli_num].fd,101,"Starting pppd\n");
	if(pppd_call(value))
	    send_reply(config->client[cli_num].fd,102,"pppd start was unsucessful\n");
	else
	    send_reply(config->client[cli_num].fd,101,"pppd started ok\n");
	return 0;
    }
}

int command_stop(config_t * config,char * cmnd_ln,int cli_num)
{
    char * value;

    /* skip the stop keyword */
    value=&cmnd_ln[4];
    while(value[0]==' ')
	value++;

    if ((value==NULL) || (strlen(value)==0))
	value="ppp0";/* default interface */

    send_reply(config->client[cli_num].fd,101,"Stoping interface %s\n",value);
    if(pppd_stop(value))
	send_reply(config->client[cli_num].fd,102,"ppp stop was unsucessful\n");
    else
	send_reply(config->client[cli_num].fd,101,"ppp stop ok\n");
    return 0;
}
